# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module contains a number of utilities for use during
setup/build/packaging that are useful to astropy as a whole.
"""

import logging
import os
import shutil
import subprocess
import sys
from collections import defaultdict

from setuptools import Extension, find_packages
from setuptools.command.build_ext import new_compiler

from ._utils import (
    abi_to_versions,
    get_limited_api_option,
    import_file,
    walk_skip_hidden,
)

__all__ = ["get_compiler", "get_extensions", "pkg_config"]

log = logging.getLogger(__name__)


def get_compiler():
    """
    Determines the compiler that will be used to build extension modules.

    Returns
    -------
    compiler : str
        The compiler option specified for the build, build_ext, or build_clib
        command; or the default compiler for the platform if none was
        specified.

    """
    return new_compiler().compiler_type


def get_extensions(srcdir="."):
    """
    Collect all extensions from Cython files and ``setup_package.py`` files.

    If numpy is importable, the numpy include path will be added to all Cython
    extensions which are automatically generated.

    This function obtains that information by iterating through all
    packages in ``srcdir`` and locating a ``setup_package.py`` module.
    This module can contain the ``get_extensions()`` function which returns
    a list of :class:`setuptools.Extension` objects.

    """
    ext_modules = []
    packages = []
    package_dir = {}

    # Use the find_packages tool to locate all packages and modules
    packages = find_packages(srcdir)

    # Update package_dir if the package lies in a subdirectory
    if srcdir != ".":
        package_dir[""] = srcdir

    for setuppkg in iter_setup_packages(srcdir, packages):
        # get_extensions must include any Cython extensions by their .pyx
        # filename.
        if hasattr(setuppkg, "get_extensions"):
            ext_modules.extend(setuppkg.get_extensions())

    # Locate any .pyx files not already specified, and add their extensions in.
    # The default include dirs include numpy to facilitate numerical work.
    includes = []
    try:
        import numpy

        includes = [numpy.get_include()]
    except ImportError:
        pass

    ext_modules.extend(get_cython_extensions(srcdir, packages, ext_modules, includes))

    # Now remove extensions that have the special name 'skip_cython', as they
    # exist Only to indicate that the cython extensions shouldn't be built
    for i, ext in reversed(list(enumerate(ext_modules))):
        if ext.name == "skip_cython":
            del ext_modules[i]

    # On Microsoft compilers, we need to pass the '/MANIFEST'
    # commandline argument.  This was the default on MSVC 9.0, but is
    # now required on MSVC 10.0, but it doesn't seem to hurt to add
    # it unconditionally.
    if get_compiler() == "msvc":
        for ext in ext_modules:
            ext.extra_link_args.append("/MANIFEST")

    if len(ext_modules) > 0:
        main_package_dir = min(packages, key=len)
        src_path = os.path.join(os.path.dirname(__file__), "src")
        shutil.copyfile(
            os.path.join(src_path, "compiler.c"),
            os.path.join(srcdir, main_package_dir, "_compiler.c"),
        )
        ext = Extension(
            main_package_dir + ".compiler_version", [os.path.join(main_package_dir, "_compiler.c")]
        )
        ext_modules.append(ext)

    # Since https://github.com/astropy/extension-helpers/pull/67,
    # extensions that used absolute paths in source names stopped working.
    # Absolute paths in source paths are undesirable but we need to
    # preserve backward-compatibility until we bump the major release,
    # so we check for the case of absolute paths and emit a deprecation
    # warning for now.
    for extension in ext_modules:
        sources = []
        fixed = []
        for source in extension.sources:
            if os.path.isabs(source):
                try:
                    source = os.path.relpath(source)
                except ValueError:
                    # In some cases it's impossible to use a relative path, for
                    # instance if the source files are on a different drive. In
                    # this case there's not much we can do so we just proceed.
                    pass
                fixed.append(source)
            sources.append(source)
        if fixed:
            log.warning(
                "Extension {} contains source files "
                "({}) that are specified using an absolute "
                "path, which will not be supported in future.".format(
                    extension.name, ", ".join(fixed)
                )
            )

        extension.sources = sources

    abi = get_limited_api_option(srcdir=srcdir)
    if abi:
        version_info, version_hex = abi_to_versions(abi)

        if version_info is None:
            raise ValueError(f"Unrecognized abi version for limited API: {abi}")

        log.info(
            f"Targeting PEP 384 limited API supporting Python >= {version_info[0], version_info[1]}"
        )

        for ext in ext_modules:
            ext.py_limited_api = True
            ext.define_macros.append(("Py_LIMITED_API", version_hex))

    return ext_modules


def iter_setup_packages(srcdir, packages):
    """A generator that finds and imports all of the ``setup_package.py``
    modules in the source packages.

    Returns
    -------
    modgen : generator
        A generator that yields (modname, mod), where `mod` is the module and
        `modname` is the module name for the ``setup_package.py`` modules.

    """

    for packagename in packages:
        package_parts = packagename.split(".")
        package_path = os.path.join(srcdir, *package_parts)
        setup_package = os.path.join(package_path, "setup_package.py")

        if os.path.isfile(setup_package):
            module = import_file(setup_package, name=packagename + ".setup_package")
            yield module


def iter_pyx_files(package_dir, package_name):
    """
    A generator that yields Cython source files (ending in '.pyx') in the
    source packages.

    Returns
    -------
    pyxgen : generator
        A generator that yields (extmod, fullfn) where `extmod` is the
        full name of the module that the .pyx file would live in based
        on the source directory structure, and `fullfn` is the path to
        the .pyx file.
    """
    for dirpath, _dirnames, filenames in walk_skip_hidden(package_dir):
        for fn in filenames:
            if fn.endswith(".pyx"):
                fullfn = os.path.join(dirpath, fn)
                # Package must match file name
                extmod = ".".join([package_name, fn[:-4]])
                yield (extmod, fullfn)

        break  # Don't recurse into subdirectories


def get_cython_extensions(srcdir, packages, prevextensions=tuple(), extincludedirs=None):
    """
    Looks for Cython files and generates Extensions if needed.

    Parameters
    ----------
    srcdir : str
        Path to the root of the source directory to search.
    prevextensions : list
        The extensions that are already defined, as a list of of
        `~setuptools.Extension` objects.  Any .pyx files already here will
        be ignored.
    extincludedirs : list or None
        Directories to include as the `include_dirs` argument to the generated
        `~setuptools.Extension` objects, as a list of strings.

    Returns
    -------
    exts : list
        The new extensions that are needed to compile all .pyx files (does not
        include any already in `prevextensions`).
    """

    # Vanilla setuptools and old versions of distribute include Cython files
    # as .c files in the sources, not .pyx, so we cannot simply look for
    # existing .pyx sources in the previous sources, but we should also check
    # for .c files with the same remaining filename. So we look for .pyx and
    # .c files, and we strip the extension.
    prevsourcepaths = []
    ext_modules = []

    for ext in prevextensions:
        for s in ext.sources:
            if s.endswith((".pyx", ".c", ".cpp")):
                sourcepath = os.path.realpath(os.path.splitext(s)[0])
                prevsourcepaths.append(sourcepath)

    for package_name in packages:
        package_parts = package_name.split(".")
        package_path = os.path.join(srcdir, *package_parts)

        for extmod, pyxfn in iter_pyx_files(package_path, package_name):
            sourcepath = os.path.realpath(os.path.splitext(pyxfn)[0])
            if sourcepath not in prevsourcepaths:
                ext_modules.append(Extension(extmod, [pyxfn], include_dirs=extincludedirs))

    return ext_modules


def pkg_config(packages, default_libraries, executable="pkg-config"):
    """
    Uses pkg-config to update a set of setuptools Extension arguments
    to include the flags necessary to link against the given packages.

    If the pkg-config lookup fails, default_libraries is applied to
    libraries.

    Parameters
    ----------
    packages : list
        The pkg-config packages to look up, as a list of strings.

    default_libraries : list
        The library names to use if the pkg-config lookup fails, a list of
        strings.

    Returns
    -------
    config : dict
        A dictionary containing keyword arguments to
        :class:`~setuptools.Extension`.  These entries include:

        - ``include_dirs``: A list of include directories
        - ``library_dirs``: A list of library directories
        - ``libraries``: A list of libraries
        - ``define_macros``: A list of macro defines
        - ``undef_macros``: A list of macros to undefine
        - ``extra_compile_args``: A list of extra arguments to pass to
          the compiler
    """

    flag_map = {
        "-I": "include_dirs",
        "-L": "library_dirs",
        "-l": "libraries",
        "-D": "define_macros",
        "-U": "undef_macros",
    }
    command = f"{executable} --libs --cflags {' '.join(packages)}"

    result = defaultdict(list)

    try:
        pipe = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
        output = pipe.communicate()[0].strip()
    except subprocess.CalledProcessError as e:
        lines = [
            (f"{executable} failed. This may cause the build to fail below."),
            f"  command: {e.cmd}",
            f"  returncode: {e.returncode}",
            f"  output: {e.output}",
        ]
        log.warning("\n".join(lines))
        result["libraries"].extend(default_libraries)
    else:
        if pipe.returncode != 0:
            lines = [
                f"pkg-config could not lookup up package(s) {', '.join(packages)}.",
                "This may cause the build to fail below.",
            ]
            log.warning("\n".join(lines))
            result["libraries"].extend(default_libraries)
        else:
            for token in output.split():
                # It's not clear what encoding the output of
                # pkg-config will come to us in.  It will probably be
                # some combination of pure ASCII (for the compiler
                # flags) and the filesystem encoding (for any argument
                # that includes directories or filenames), but this is
                # just conjecture, as the pkg-config documentation
                # doesn't seem to address it.
                arg = token[:2].decode("ascii")
                value = token[2:].decode(sys.getfilesystemencoding())
                if arg in flag_map:
                    if arg == "-D":
                        value = tuple(value.split("=", 1))
                    result[flag_map[arg]].append(value)
                else:
                    result["extra_compile_args"].append(value)

    return result



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import re
import sys
import tempfile
from configparser import ConfigParser
from importlib import machinery as import_machinery
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


__all__ = ["write_if_different", "import_file", "get_limited_api_option", "abi_to_versions"]


if sys.platform == "win32":
    import ctypes

    def _has_hidden_attribute(filepath):
        """
        Returns True if the given filepath has the hidden attribute on
        MS-Windows.  Based on a post here:
        http://stackoverflow.com/questions/284115/cross-platform-hidden-file-detection
        """
        if isinstance(filepath, bytes):
            filepath = filepath.decode(sys.getfilesystemencoding())
        try:
            attrs = ctypes.windll.kernel32.GetFileAttributesW(filepath)
            assert attrs != -1
            result = bool(attrs & 2)
        except (AttributeError, AssertionError):
            result = False
        return result

else:

    def _has_hidden_attribute(filepath):
        return False


def is_path_hidden(filepath):
    """
    Determines if a given file or directory is hidden.

    Parameters
    ----------
    filepath : str
        The path to a file or directory

    Returns
    -------
    hidden : bool
        Returns `True` if the file is hidden
    """

    name = os.path.basename(os.path.abspath(filepath))
    if isinstance(name, bytes):
        is_dotted = name.startswith(b".")
    else:
        is_dotted = name.startswith(".")
    return is_dotted or _has_hidden_attribute(filepath)


def walk_skip_hidden(top, onerror=None, followlinks=False):
    """
    A wrapper for `os.walk` that skips hidden files and directories.

    This function does not have the parameter `topdown` from
    `os.walk`: the directories must always be recursed top-down when
    using this function.

    See also
    --------
    os.walk : For a description of the parameters
    """

    for root, dirs, files in os.walk(top, topdown=True, onerror=onerror, followlinks=followlinks):
        # These lists must be updated in-place so os.walk will skip
        # hidden directories
        dirs[:] = [d for d in dirs if not is_path_hidden(d)]
        files[:] = [f for f in files if not is_path_hidden(f)]
        yield root, dirs, files


def write_if_different(filename, data):
    """
    Write ``data`` to ``filename``, if the content of the file is different.

    This can be useful if e.g. generating ``.c`` or ``.h`` files, to make sure
    that Python does not re-build unchanged files.

    Parameters
    ----------
    filename : str or `pathlib.Path`
        The file name to be written to.
    data : bytes
        The data to be written to ``filename``.
    """

    filepath = Path(filename)

    assert isinstance(data, bytes)

    if filepath.exists():
        original_data = filepath.read_bytes()
    else:
        original_data = None

    if original_data != data:
        filepath.write_bytes(data)


def import_file(filename, name=None):
    """
    Imports a module from a single file without importing the package that
    the file is in.

    This is useful for cases where a file needs to be imported from
    ``setup_package.py`` files without importing the parent package. The
    returned module will have the optional ``name`` if given, or else a name
    generated from the filename.
    """
    # Specifying a traditional dot-separated fully qualified name here
    # results in a number of "Parent module '...' not found while
    # handling absolute import" warnings.  Using the same name, the
    # namespaces of the modules get merged together.  So, this
    # generates an underscore-separated name which is more likely to
    # be unique, and it doesn't really matter because the name isn't
    # used directly here anyway.

    filepath = Path(filename)

    if name is None:
        name = "_".join(filepath.resolve().with_suffix("").parts[1:])

    if not filepath.exists():
        raise ImportError(f"Could not import file {filepath}")

    loader = import_machinery.SourceFileLoader(name, str(filepath))
    spec = spec_from_file_location(name, str(filepath))
    mod = module_from_spec(spec)
    loader.exec_module(mod)

    return mod


def get_limited_api_option(srcdir):
    """
    Checks setup.cfg and pyproject.toml files in the current directory
    for the py_limited_api setting
    """

    py_limited_api = os.environ.get("EXTENSION_HELPERS_PY_LIMITED_API")

    if py_limited_api is not None:

        if "DIST_EXTRA_CONFIG" in os.environ:
            raise ValueError(
                "Cannot use EXTENSION_HELPERS_PY_LIMITED_API if DIST_EXTRA_CONFIG is already defined"
            )

        dist_extra_config_filename = tempfile.mktemp()
        with open(dist_extra_config_filename, "w") as f:
            f.write(f"[bdist_wheel]\npy_limited_api={py_limited_api}")
        os.environ["DIST_EXTRA_CONFIG"] = dist_extra_config_filename
        return py_limited_api

    srcdir = Path(srcdir)

    setup_cfg = srcdir / "setup.cfg"

    if setup_cfg.exists():
        cfg = ConfigParser()
        cfg.read(setup_cfg)
        if cfg.has_option("bdist_wheel", "py_limited_api"):
            return cfg.get("bdist_wheel", "py_limited_api")

    pyproject = srcdir / "pyproject.toml"
    if pyproject.exists():
        with pyproject.open("rb") as f:
            pyproject_cfg = tomllib.load(f)
            if (
                "tool" in pyproject_cfg
                and "distutils" in pyproject_cfg["tool"]
                and "bdist_wheel" in pyproject_cfg["tool"]["distutils"]
                and "py-limited-api" in pyproject_cfg["tool"]["distutils"]["bdist_wheel"]
            ):
                return pyproject_cfg["tool"]["distutils"]["bdist_wheel"]["py-limited-api"]


def _abi_to_version_info(abi):
    match = re.fullmatch(r"^cp(\d)(\d+)$", abi)
    if match is None:
        return None
    else:
        return int(match[1]), int(match[2])


def _version_info_to_version_hex(major=0, minor=0):
    """Returns a PY_VERSION_HEX for {major}.{minor).0"""
    return f"0x{major:02X}{minor:02X}0000"


def abi_to_versions(abi):
    version_info = _abi_to_version_info(abi)
    if version_info is None:
        return None, None
    else:
        return version_info, _version_info_to_version_hex(*version_info)



import importlib
import os
import subprocess
import sys
import uuid
from textwrap import dedent

import pytest

from .._setup_helpers import get_compiler, get_extensions
from . import cleanup_import, run_setup

if sys.version_info >= (3, 11):
    from contextlib import chdir
else:
    from .py311_backports import chdir

extension_helpers_PATH = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", "..")
)  # noqa


def teardown_module(module):
    # Remove file generated by test_generate_openmp_enabled_py but
    # somehow needed in test_cython_autoextensions
    tmpfile = "openmp_enabled.py"
    if os.path.exists(tmpfile):
        os.remove(tmpfile)


POSSIBLE_COMPILERS = ["unix", "msvc", "bcpp", "cygwin", "mingw32"]


def test_get_compiler():
    assert get_compiler() in POSSIBLE_COMPILERS


def _extension_test_package(
    tmp_path,
    request=None,
    extension_type="c",
    include_numpy=False,
    include_setup_py=True,
):
    """Creates a simple test package with an extension module."""

    test_pkg = tmp_path / "test_pkg"
    os.makedirs(test_pkg / "helpers_test_package")
    (test_pkg / "helpers_test_package" / "__init__.py").touch()

    # TODO: It might be later worth making this particular test package into a
    # reusable fixture for other build_ext tests

    if extension_type in ("c", "both"):
        # A minimal C extension for testing
        (test_pkg / "helpers_test_package" / "unit01.c").write_text(
            dedent(
                """\
            #include <Python.h>

            static struct PyModuleDef moduledef = {
                PyModuleDef_HEAD_INIT,
                "unit01",
                NULL,
                -1,
                NULL
            };
            PyMODINIT_FUNC
            PyInit_unit01(void) {
                return PyModule_Create(&moduledef);
            }
        """
            )
        )

    if extension_type in ("pyx", "both"):
        # A minimal Cython extension for testing
        (test_pkg / "helpers_test_package" / "unit02.pyx").write_text(
            dedent(
                """\
            print("Hello cruel angel.")
        """
            )
        )

    if extension_type == "c":
        extensions = ["unit01.c"]
    elif extension_type == "pyx":
        extensions = ["unit02.pyx"]
    elif extension_type == "both":
        extensions = ["unit01.c", "unit02.pyx"]

    include_dirs = ["numpy"] if include_numpy else []

    extensions_list = [
        f"Extension('helpers_test_package.{os.path.splitext(extension)[0]}', "
        f"[join('helpers_test_package', '{extension}')], "
        f"{include_dirs=})"
        for extension in extensions
    ]

    (test_pkg / "helpers_test_package" / "setup_package.py").write_text(
        dedent(
            """\
        from setuptools import Extension
        from os.path import join
        def get_extensions():
            return [{}]
    """.format(
                ", ".join(extensions_list)
            )
        )
    )

    if include_setup_py:
        (test_pkg / "setup.py").write_text(
            dedent(
                f"""\
            import sys
            from os.path import join
            from setuptools import setup, find_packages
            sys.path.insert(0, r'{extension_helpers_PATH}')
            from extension_helpers import get_extensions

            setup(
                name='helpers_test_package',
                version='0.1',
                packages=find_packages(),
                ext_modules=get_extensions()
            )
        """
            )
        )

    if "" in sys.path:
        sys.path.remove("")

    sys.path.insert(0, "")

    def finalize():
        cleanup_import("helpers_test_package")

    if request:
        request.addfinalizer(finalize)

    return test_pkg


@pytest.fixture
def extension_test_package(tmp_path, request):
    return _extension_test_package(tmp_path, request, extension_type="both")


@pytest.fixture
def c_extension_test_package(tmp_path, request):
    # Check whether numpy is installed in the test environment
    has_numpy = bool(importlib.util.find_spec("numpy"))
    return _extension_test_package(tmp_path, request, extension_type="c", include_numpy=has_numpy)


@pytest.fixture
def pyx_extension_test_package(tmp_path, request):
    return _extension_test_package(tmp_path, request, extension_type="pyx")


def test_cython_autoextensions(tmp_path):
    """
    Regression test for https://github.com/astropy/astropy-helpers/pull/19

    Ensures that Cython extensions in sub-packages are discovered and built
    only once.
    """

    # Make a simple test package

    test_pkg = tmp_path / "test_pkg"
    os.makedirs(test_pkg / "yoda" / "luke")
    (test_pkg / "yoda" / "__init__.py").touch()
    (test_pkg / "yoda" / "luke" / "__init__.py").touch()
    (test_pkg / "yoda" / "luke" / "dagobah.pyx").write_text("""def testfunc(): pass""")

    # Required, currently, for get_extensions to work
    ext_modules = get_extensions(str(test_pkg))

    assert len(ext_modules) == 2
    assert ext_modules[0].name == "yoda.luke.dagobah"


def test_compiler_module(capsys, c_extension_test_package):
    """
    Test ensuring that the compiler module is built and installed for packages
    that have extension modules.
    """

    test_pkg = c_extension_test_package
    install_temp = test_pkg / "install_temp"
    os.mkdir(install_temp)

    with chdir(test_pkg):
        # This is one of the simplest ways to install just a package into a
        # test directory
        run_setup(
            "setup.py",
            [
                "install",
                "--single-version-externally-managed",
                f"--install-lib={install_temp}",
                "--record={}".format(install_temp / "record.txt"),
            ],
        )

    with chdir(install_temp):
        import helpers_test_package

        # Make sure we imported the helpers_test_package package from the correct place
        dirname = os.path.abspath(os.path.dirname(helpers_test_package.__file__))
        assert dirname == str(install_temp / "helpers_test_package")

        import helpers_test_package.compiler_version

        assert helpers_test_package.compiler_version != "unknown"


@pytest.mark.parametrize("use_extension_helpers", [None, False, True])
@pytest.mark.parametrize("pyproject_use_helpers", [None, False, True])
def test_no_setup_py(tmp_path, use_extension_helpers, pyproject_use_helpers):
    """
    Test that makes sure that extension-helpers can be enabled without a
    setup.py file.
    """

    package_name = "helpers_test_package_" + str(uuid.uuid4()).replace("-", "_")

    test_pkg = tmp_path / "test_pkg"
    os.makedirs(test_pkg / package_name)
    (test_pkg / package_name / "__init__.py").touch()

    simple_c = test_pkg / package_name / "simple.c"

    simple_c.write_text(
        dedent(
            """\
        #include <Python.h>

        static struct PyModuleDef moduledef = {
            PyModuleDef_HEAD_INIT,
            "simple",
            NULL,
            -1,
            NULL
        };
        PyMODINIT_FUNC
        PyInit_simple(void) {
            return PyModule_Create(&moduledef);
        }
    """
        )
    )

    (test_pkg / package_name / "setup_package.py").write_text(
        dedent(
            f"""\
        from setuptools import Extension
        from os.path import join
        def get_extensions():
            return [Extension('{package_name}.simple', [join('{package_name}', 'simple.c')])]
        """
        )
    )

    if use_extension_helpers is None:
        (test_pkg / "setup.cfg").write_text(
            dedent(
                f"""\
            [metadata]
            name = {package_name}
            version = 0.1

            [options]
            packages = find:
        """
            )
        )
    else:
        (test_pkg / "setup.cfg").write_text(
            dedent(
                f"""\
            [metadata]
            name = {package_name}
            version = 0.1

            [options]
            packages = find:

            [extension-helpers]
            use_extension_helpers = {str(use_extension_helpers).lower()}
        """
            )
        )

    if pyproject_use_helpers is None:
        (test_pkg / "pyproject.toml").write_text(
            dedent(
                """\
            [build-system]
            requires = ["setuptools>=43.0.0",
                        "wheel"]
            build-backend = 'setuptools.build_meta'
        """
            )
        )
    else:
        (test_pkg / "pyproject.toml").write_text(
            dedent(
                f"""\
            [build-system]
            requires = ["setuptools>=43.0.0",
                        "wheel"]
            build-backend = 'setuptools.build_meta'

            [tool.extension-helpers]
            use_extension_helpers = {str(pyproject_use_helpers).lower()}
        """
            )
        )

    install_temp = test_pkg / "install_temp"
    os.mkdir(install_temp)

    with chdir(test_pkg):
        # NOTE: we disable build isolation as we need to pick up the current
        # developer version of extension-helpers
        subprocess.call(
            [
                sys.executable,
                "-m",
                "pip",
                "install",
                ".",
                "--no-build-isolation",
                f"--target={install_temp}",
            ]
        )

    if "" in sys.path:
        sys.path.remove("")

    sys.path.insert(0, "")

    with chdir(install_temp):
        importlib.import_module(package_name)

        if use_extension_helpers or (use_extension_helpers is None and pyproject_use_helpers):
            compiler_version_mod = importlib.import_module(package_name + ".compiler_version")
            assert compiler_version_mod.compiler != "unknown"
        else:
            try:
                importlib.import_module(package_name + ".compiler_version")
            except ImportError:
                pass
            else:
                raise AssertionError(package_name + ".compiler_version should not exist")


@pytest.mark.parametrize("pyproject_use_helpers", [None, False, True])
def test_only_pyproject(tmp_path, pyproject_use_helpers):
    """
    Test that makes sure that extension-helpers can be enabled without a
    setup.py and without a setup.cfg file.
    """

    pytest.importorskip("setuptools", minversion="62.0")

    package_name = "helpers_test_package_" + str(uuid.uuid4()).replace("-", "_")

    test_pkg = tmp_path / "test_pkg"
    os.makedirs(test_pkg / package_name)
    (test_pkg / package_name / "__init__.py").touch()
    simple_pyx = test_pkg / package_name / "simple.pyx"
    simple_pyx.write_text(
        dedent(
            """\
        def test():
            pass
    """
        )
    )

    if pyproject_use_helpers is None:
        extension_helpers_option = ""
    else:
        extension_helpers_option = dedent(
            f"""
        [tool.extension-helpers]
        use_extension_helpers = {str(pyproject_use_helpers).lower()}
        """
        )

    buildtime_requirements = ["setuptools>=43.0.0", "wheel", "Cython"]
    (test_pkg / "pyproject.toml").write_text(
        dedent(
            f"""\
            [project]
            name = "{package_name}"
            version = "0.1"

            [tool.setuptools.packages]
            find = {{namespaces = false}}

            [build-system]
            requires = [{', '.join(f'"{_}"' for _ in buildtime_requirements)}]
            build-backend = 'setuptools.build_meta'

            """
        )
        + extension_helpers_option
    )

    install_temp = test_pkg / "install_temp"
    os.mkdir(install_temp)

    with chdir(test_pkg):
        # NOTE: we disable build isolation as we need to pick up the current
        # developer version of extension-helpers
        # In order to do so, we need to ensure that build-time dependencies are
        # installed first
        cmd1 = [
            sys.executable,
            "-m",
            "pip",
            "install",
            *buildtime_requirements,
            f"--target={install_temp}",
        ]
        subprocess.call(cmd1)

        cmd2 = [
            sys.executable,
            "-m",
            "pip",
            "install",
            ".",
            "--no-build-isolation",
            f"--target={install_temp}",
        ]
        subprocess.call(cmd2)

    if "" in sys.path:
        sys.path.remove("")

    sys.path.insert(0, "")

    with chdir(install_temp):
        importlib.import_module(package_name)

        if pyproject_use_helpers:
            compiler_version_mod = importlib.import_module(package_name + ".compiler_version")
            assert compiler_version_mod.compiler != "unknown"
        else:
            try:
                importlib.import_module(package_name + ".compiler_version")
            except ImportError:
                pass
            else:
                raise AssertionError(package_name + ".compiler_version should not exist")


# Tests to make sure that limited API support works correctly


@pytest.mark.parametrize("config", ("setup.cfg", "pyproject.toml"))
@pytest.mark.parametrize("envvar", (False, True))
@pytest.mark.parametrize("limited_api", (None, "cp310"))
@pytest.mark.parametrize("extension_type", ("c", "pyx", "both"))
def test_limited_api(tmp_path, config, envvar, limited_api, extension_type):

    if sys.version_info < (3, 11):
        pytest.skip(
            "This test requires setuptools>=65.4 which is only available for Python 3.11 and later"
        )

    package = _extension_test_package(
        tmp_path, extension_type=extension_type, include_numpy=True, include_setup_py=False
    )

    if config == "setup.cfg":

        setup_cfg = dedent(
            """\
            [metadata]
            name = helpers_test_package
            version = 0.1

            [options]
            packages = find:

            [extension-helpers]
            use_extension_helpers = true
        """
        )

        if limited_api and not envvar:
            setup_cfg += f"\n[bdist_wheel]\npy_limited_api={limited_api}"
        elif envvar:
            # Make sure if we are using the environment variable that it takes
            # precedence over this setting (this only works for setup.cfg)
            setup_cfg += "\n[bdist_wheel]\npy_limited_api=cp35"

        (package / "setup.cfg").write_text(setup_cfg)

        # Still require a minimal pyproject.toml file if no setup.py file

        (package / "pyproject.toml").write_text(
            dedent(
                """
            [build-system]
            requires = ["setuptools>=43.0.0",
                        "wheel"]
            build-backend = 'setuptools.build_meta'

            [tool.extension-helpers]
            use_extension_helpers = true
        """
            )
        )

    elif config == "pyproject.toml":

        pyproject_toml = dedent(
            """\
            [build-system]
            requires = ["setuptools>=43.0.0",
                        "wheel"]
            build-backend = 'setuptools.build_meta'

            [project]
            name = "helpers_test_package"
            version = "0.1"

            [tool.setuptools.packages]
            find = {namespaces = false}

            [tool.extension-helpers]
            use_extension_helpers = true
            """
        )

        if limited_api and not envvar:
            pyproject_toml += f'\n[tool.distutils.bdist_wheel]\npy-limited-api = "{limited_api}"'

        (package / "pyproject.toml").write_text(pyproject_toml)

    env = os.environ.copy()

    if envvar:
        if limited_api:
            env["EXTENSION_HELPERS_PY_LIMITED_API"] = limited_api
        else:
            env["EXTENSION_HELPERS_PY_LIMITED_API"] = ""

    with chdir(package):
        subprocess.run(
            [sys.executable, "-m", "build", "--wheel", "--no-isolation"], env=env, check=True
        )

    wheels = os.listdir(package / "dist")

    assert len(wheels) == 1
    assert ("abi3" in wheels[0]) == (limited_api is not None)


def test_limited_api_invalid_abi(tmp_path, capsys):

    package = _extension_test_package(
        tmp_path, extension_type="c", include_numpy=True, include_setup_py=False
    )

    (package / "setup.cfg").write_text(
        dedent(
            """\
        [metadata]
        name = helpers_test_package
        version = 0.1

        [options]
        packages = find:

        [extension-helpers]
        use_extension_helpers = true

        [bdist_wheel]
        py_limited_api=invalid
    """
        )
    )

    (package / "pyproject.toml").write_text(
        dedent(
            """
    [build-system]
    requires = ["setuptools>=43.0.0",
                "wheel"]
    build-backend = 'setuptools.build_meta'
    """
        )
    )

    with chdir(package):
        result = subprocess.run(
            [sys.executable, "-m", "build", "--wheel", "--no-isolation"], stderr=subprocess.PIPE
        )

    assert result.stderr.strip().endswith(
        b"ValueError: Unrecognized abi version for limited API: invalid"
    )
